from setuptools import setup, find_packages
from standard_j import __version__

setup(
    name="standard_j",
    version=__version__,
    author="J7W539",
    author_email="j7w539@gmail.com",
    url="https://github.com/J7W539/standard-j",
    packages=find_packages(),
)
